@@ -11,45 +11,59 @@ from io import BytesIO
 from itertools import zip_longest
 from pathlib import Path
 from shutil import copyfileobj
-from tempfile import NamedTemporaryFile
-from zlib import decompress, error as ZlibError
+from zlib import decompress
 
-from .. import Array, Dictionary, Name, Object, PdfError, Stream
+from PIL import Image, ImageCms
+from PIL.TiffTags import TAGS_V2 as TIFF_TAGS
+
+from pikepdf import (
+    Array,
+    Dictionary,
+    Name,
+    Object,
+    PdfError,
+    Stream,
+    StreamDecodeLevel,
+    String,
+    jbig2,
+)
 
 
 class DependencyError(Exception):
-    pass
+    "A third party dependency is needed to extract images of this type."
 
 
 class UnsupportedImageTypeError(Exception):
-    pass
+    "This image is formatted in a way pikepdf does not supported."
 
 
 class NotExtractableError(Exception):
-    pass
+    "Indicates that an image cannot be directly extracted."
+
+
+class InvalidPdfImageError(Exception):
+    "This image is not valid according to the PDF 1.7 specification."
 
 
 def array_str(value):
-    if isinstance(value, (list, Array)):
-        return [str(item) for item in value]
-    if isinstance(value, Name):
-        return [str(value)]
-    raise NotImplementedError(value)
+    """Simplify pikepdf objects to array of str. Keep Streams intact."""
 
+    def _array_str(item):
+        if isinstance(item, (list, Array)):
+            return [_array_str(subitem) for subitem in item]
+        elif isinstance(item, (Stream, bytes, int)):
+            return item
+        elif isinstance(item, (Name, str)):
+            return str(item)
+        elif isinstance(item, (String)):
+            return bytes(item)
+        else:
+            raise NotImplementedError(value)
 
-def array_str_colorspace(value):
-    if isinstance(value, (list, Array)):
-        items = [item for item in value]
-        if len(items) == 4 and items[0] == '/Indexed':
-            result = [str(items[n]) for n in range(3)]
-            result.append(bytes(items[3]))
-            return result
-        if len(items) == 2 and items[0] == '/ICCBased':
-            result = [str(items[0]), items[1]]
-            return result
-        return array_str(items)
-
-    return array_str(value)
+    result = _array_str(value)
+    if not isinstance(result, list):
+        result = [result]
+    return result
 
 
 def dict_or_array_dict(value):
@@ -58,7 +72,7 @@ def dict_or_array_dict(value):
     if isinstance(value, Dictionary):
         return [value.as_dict()]
     if isinstance(value, Array):
-        return [v.as_list() for v in value]
+        return [v for v in value.as_list()]
     raise NotImplementedError(value)
 
 
@@ -74,11 +88,12 @@ def metadata_from_obj(obj, name, type_, default):
 
 class PdfImageBase(ABC):
 
-    SIMPLE_COLORSPACES = ('/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray')
+    SIMPLE_COLORSPACES = {'/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray'}
+    MAIN_COLORSPACES = SIMPLE_COLORSPACES | {'/DeviceCMYK', '/CalCMYK', '/ICCBased'}
 
     @abstractmethod
     def _metadata(self, name, type_, default):
-        pass
+        ...
 
     @property
     def width(self):
@@ -103,7 +118,7 @@ class PdfImageBase(ABC):
     @property
     def _colorspaces(self):
         """Colorspace (low-level)"""
-        return self._metadata('ColorSpace', array_str_colorspace, [])
+        return self._metadata('ColorSpace', array_str, [])
 
     @property
     def filters(self):
@@ -121,15 +136,14 @@ class PdfImageBase(ABC):
         if self.image_mask:
             return None  # Undefined for image masks
         if self._colorspaces:
-            if self._colorspaces[0] in self.SIMPLE_COLORSPACES:
+            if self._colorspaces[0] in self.MAIN_COLORSPACES:
                 return self._colorspaces[0]
-            if self._colorspaces[0] in ('/DeviceCMYK', '/ICCBased'):
-                return self._colorspaces[0]
-            if (
-                self._colorspaces[0] == '/Indexed'
-                and self._colorspaces[1] in self.SIMPLE_COLORSPACES
-            ):
-                return self._colorspaces[1]
+            if self._colorspaces[0] == '/Indexed':
+                subspace = self._colorspaces[1]
+                if isinstance(subspace, str) and subspace in self.MAIN_COLORSPACES:
+                    return subspace
+                if isinstance(subspace, list) and subspace[0] == '/ICCBased':
+                    return subspace[0]
         raise NotImplementedError(
             "not sure how to get colorspace: " + repr(self._colorspaces)
         )
@@ -144,12 +158,12 @@ class PdfImageBase(ABC):
     @property
     @abstractmethod
     def is_inline(self):
-        pass
+        ...
 
     @property
     @abstractmethod
     def icc(self):
-        pass
+        ...
 
     @property
     def indexed(self):
@@ -161,6 +175,21 @@ class PdfImageBase(ABC):
         """Size of image as (width, height)"""
         return self.width, self.height
 
+    def _approx_mode_from_icc(self):
+        if self.indexed:
+            icc_profile = self._colorspaces[1][1]
+        else:
+            icc_profile = self._colorspaces[1]
+        icc_profile_nchannels = int(icc_profile['/N'])
+
+        if icc_profile_nchannels == 1:
+            return 'L'
+
+        # Multiple channels, need to open the profile and look
+        mode_from_xcolor_space = {'RGB ': 'RGB', 'CMYK': 'CMYK'}
+        xcolor_space = self.icc.profile.xcolor_space
+        return mode_from_xcolor_space.get(xcolor_space, '')
+
     @property
     def mode(self):
         """``PIL.Image.mode`` equivalent for this image, where possible
@@ -183,18 +212,15 @@ class PdfImageBase(ABC):
                 m = 'CMYK'
             elif self.colorspace == '/ICCBased':
                 try:
-                    icc_profile = self._colorspaces[1]
-                    icc_profile_nchannels = int(icc_profile['/N'])
-                    if icc_profile_nchannels == 1:
-                        m = 'L'
-                    elif icc_profile_nchannels == 3:
-                        m = 'RGB'
-                    elif icc_profile_nchannels == 4:
-                        m = 'CMYK'
-                except (ValueError, TypeError):
-                    pass
+                    m = self._approx_mode_from_icc()
+                except (ValueError, TypeError) as e:
+                    raise NotImplementedError(
+                        "Not sure how to handle PDF image of this type"
+                    ) from e
         if m == '':
-            raise NotImplementedError("Not sure how to handle PDF image of this type")
+            raise NotImplementedError(
+                "Not sure how to handle PDF image of this type"
+            ) from None
         return m
 
     @property
@@ -221,25 +247,30 @@ class PdfImageBase(ABC):
 
         if not self.indexed:
             return None
-        _idx, base, hival, lookup = None, None, None, None
         try:
-            _idx, base, hival, lookup = self._colorspaces
+            _idx, base, _hival, lookup = self._colorspaces
         except ValueError as e:
             raise ValueError('Not sure how to interpret this palette') from e
-        base = str(base)
-        hival = int(hival)
+        if self.icc:
+            base = str(base[0])
+        else:
+            base = str(base)
         lookup = bytes(lookup)
-        if not base in self.SIMPLE_COLORSPACES:
+        if not base in self.MAIN_COLORSPACES:
             raise NotImplementedError("not sure how to interpret this palette")
         if base == '/DeviceRGB':
             base = 'RGB'
         elif base == '/DeviceGray':
             base = 'L'
+        elif base == '/DeviceCMYK':
+            base = 'CMYK'
+        elif base == '/ICCBased':
+            base = self._approx_mode_from_icc()
         return base, lookup
 
     @abstractmethod
     def as_pil_image(self):
-        pass
+        ...
 
     @staticmethod
     def _unstack_compression(buffer, filters):
@@ -257,10 +288,7 @@ class PdfImageBase(ABC):
         """
         data = memoryview(buffer)
         while len(filters) > 1 and filters[0] == '/FlateDecode':
-            try:
-                data = decompress(data)
-            except ZlibError as e:
-                raise UnsupportedImageTypeError() from e
+            data = decompress(data)
             filters = filters[1:]
         return data, filters
 
@@ -297,6 +325,11 @@ class PdfImage(PdfImageBase):
         self.obj = obj
         self._icc = None
 
+    def __eq__(self, other):
+        if not isinstance(other, PdfImageBase):
+            return NotImplemented
+        return self.obj == other.obj
+
     @classmethod
     def _from_pil_image(cls, *, pdf, page, name, image):  # pragma: no cover
         """Insert a PIL image into a PDF (rudimentary)
@@ -333,6 +366,15 @@ class PdfImage(PdfImageBase):
         """``False`` for image XObject"""
         return False
 
+    @property
+    def _iccstream(self):
+        if self.colorspace == '/ICCBased':
+            if not self.indexed:
+                return self._colorspaces[1]
+            assert isinstance(self._colorspaces[1], list)
+            return self._colorspaces[1][1]
+        raise NotImplementedError("Don't know how to find ICC stream for image")
+
     @property
     def icc(self):
         """If an ICC profile is attached, return a Pillow object that describe it.
@@ -342,15 +384,20 @@ class PdfImage(PdfImageBase):
         Returns:
             PIL.ImageCms.ImageCmsProfile
         """
-        from PIL import ImageCms
-
-        if self.colorspace != '/ICCBased':
+        if self.colorspace not in ('/ICCBased', '/Indexed'):
             return None
         if not self._icc:
-            iccstream = self._colorspaces[1]
+            iccstream = self._iccstream
             iccbuffer = iccstream.get_stream_buffer()
             iccbytesio = BytesIO(iccbuffer)
-            self._icc = ImageCms.ImageCmsProfile(iccbytesio)
+            try:
+                self._icc = ImageCms.ImageCmsProfile(iccbytesio)
+            except OSError as e:
+                if str(e) == 'cannot open profile from string':
+                    # ICC profile is corrupt
+                    raise UnsupportedImageTypeError(
+                        "ICC profile corrupt or not readable"
+                    ) from e
         return self._icc
 
     def _extract_direct(self, *, stream):
@@ -387,8 +434,10 @@ class PdfImage(PdfImageBase):
 
         if filters == ['/CCITTFaxDecode']:
             if self.colorspace == '/ICCBased':
-                raise UnsupportedImageTypeError("Cannot direct-extract CCITT + ICC")
-            stream.write(self._generate_ccitt_header(data))
+                icc = self._iccstream.read_bytes()
+            else:
+                icc = None
+            stream.write(self._generate_ccitt_header(data, icc=icc))
             stream.write(data)
             return '.tif'
         elif filters == ['/DCTDecode'] and (
@@ -400,8 +449,6 @@ class PdfImage(PdfImageBase):
         raise NotExtractableError()
 
     def _extract_transcoded(self):
-        from PIL import Image
-
         im = None
         if self.mode == 'RGB' and self.bits_per_component == 8:
             # No point in accessing the buffer here, size qpdf decodes to 3-byte
@@ -419,17 +466,36 @@ class PdfImage(PdfImageBase):
                     im.putpalette(palette, rawmode=base_mode)
                 else:
                     raise NotImplementedError('palette with ' + base_mode)
-        elif self.mode == '1' and self.bits_per_component == 1:
-            data = self.read_bytes()
-            im = Image.frombytes('1', self.size, data)
-
-        elif self.mode == 'P' and self.bits_per_component == 1:
-            data = self.read_bytes()
-            im = Image.frombytes('1', self.size, data)
+        elif self.bits_per_component == 1:
+            if self.filters and self.filters[0] == '/JBIG2Decode':
+                if not jbig2.jbig2dec_available():
+                    raise DependencyError(
+                        "jbig2dec - not installed or installed version is too old "
+                        "(older than version 0.15)"
+                    )
+                jbig2_globals_obj = self.filter_decodeparms[0][1].get('/JBIG2Globals')
+                im = jbig2.extract_jbig2(self.obj, jbig2_globals_obj)
+            else:
+                data = self.read_bytes()
+                im = Image.frombytes('1', self.size, data)
+        else:
+            raise UnsupportedImageTypeError(repr(self))
 
+        if self.mode == 'P' and self.bits_per_component == 1:
+            # Fix paletted 1-bit images
             base_mode, palette = self.palette
-            if not (palette == b'\x00\x00\x00\xff\xff\xff' or palette == b'\x00\xff'):
-                raise NotImplementedError('monochrome image with nontrivial palette')
+            if base_mode == 'RGB' and palette != b'\x00\x00\x00\xff\xff\xff':
+                im = im.convert('P')
+                im.putpalette(palette, rawmode=base_mode)
+                gp = im.getpalette()
+                gp[765:768] = gp[3:6]  # work around Pillow bug
+                im.putpalette(gp)
+            elif base_mode == 'L' and palette != b'\x00\xff':
+                im = im.convert('P')
+                im.putpalette(palette, rawmode=base_mode)
+                gp = im.getpalette()
+                gp[255] = gp[1]  # work around Pillow bug
+                im.putpalette(gp)
 
         if self.colorspace == '/ICCBased':
             im.info['icc_profile'] = self.icc.tobytes()
@@ -437,18 +503,13 @@ class PdfImage(PdfImageBase):
         return im
 
     def _extract_to_stream(self, *, stream):
-        """Attempt to extract the image directly to a usable image file
+        """Attempt to extract the image to a stream
 
         If possible, the compressed data is extracted and inserted into
         a compressed image file format without transcoding the compressed
         content. If this is not possible, the data will be decompressed
         and extracted to an appropriate format.
 
-        Because it is not known until attempted what image format will be
-        extracted, users should not assume what format they are getting back.
-        When saving the image to a file, use a temporary filename, and then
-        rename the file to its final name based on the returned file extension.
-
         Args:
             stream: Writable stream to write data to
 
@@ -521,13 +582,13 @@ class PdfImage(PdfImageBase):
             copyfileobj(bio, target)
         return str(filepath)
 
-    def read_bytes(self):
+    def read_bytes(self, decode_level=StreamDecodeLevel.specialized):
         """Decompress this image and return it as unencoded bytes"""
-        return self.obj.read_bytes()
+        return self.obj.read_bytes(decode_level=decode_level)
 
-    def get_stream_buffer(self):
+    def get_stream_buffer(self, decode_level=StreamDecodeLevel.specialized):
         """Access this image with the buffer protocol"""
-        return self.obj.get_stream_buffer()
+        return self.obj.get_stream_buffer(decode_level=decode_level)
 
     def as_pil_image(self):
         """Extract the image as a Pillow Image, using decompression as necessary
@@ -535,8 +596,6 @@ class PdfImage(PdfImageBase):
         Returns:
             PIL.Image.Image
         """
-        from PIL import Image
-
         try:
             bio = BytesIO()
             self._extract_direct(stream=bio)
@@ -551,7 +610,7 @@ class PdfImage(PdfImageBase):
 
         return im
 
-    def _generate_ccitt_header(self, data):
+    def _generate_ccitt_header(self, data, icc=None):
         """Construct a CCITT G3 or G4 header from the PDF metadata"""
         # https://stackoverflow.com/questions/2641770/
         # https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf
@@ -570,7 +629,7 @@ class PdfImage(PdfImageBase):
         elif k > 0:
             ccitt_group = 3  # Group 3 2-D
         else:
-            ccitt_group = 2  # CCITT 1-D
+            ccitt_group = 2  # Group 3 1-D
         black_is_one = self.decode_parms[0].get("/BlackIs1", False)
         # TIFF spec says:
         # use 0 for white_is_zero (=> black is 1)
@@ -579,34 +638,69 @@ class PdfImage(PdfImageBase):
         photometry = 1 if black_is_one else 0
 
         img_size = len(data)
-        tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H' + 'HHLL' * 8 + 'L'
-        # fmt: off
+        tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H'
+
+        tag_keys = {tag.name: key for key, tag in TIFF_TAGS.items()}
+        ifd_struct = '<HHLL'
+
+        if icc is None:
+            icc = b''
+
+        ifds = []
+
+        def header_length(ifd_count):
+            return (
+                struct.calcsize(tiff_header_struct)
+                + struct.calcsize(ifd_struct) * ifd_count
+                + 4
+            )
+
+        def add_ifd(tag_name, data, count=1):
+            key = tag_keys[tag_name]
+            typecode = TIFF_TAGS[key].type
+            ifds.append((key, typecode, count, data))
+
+        image_offset = None
+        add_ifd('ImageWidth', self.width)
+        add_ifd('ImageLength', self.height)
+        add_ifd('BitsPerSample', 1)
+        add_ifd('Compression', ccitt_group)
+        add_ifd('PhotometricInterpretation', int(photometry))
+        add_ifd('StripOffsets', lambda: image_offset)
+        add_ifd('RowsPerStrip', self.height)
+        add_ifd('StripByteCounts', img_size)
+
+        icc_offset = 0
+        if icc:
+            add_ifd('ICCProfile', lambda: icc_offset, count=len(icc))
+
+        icc_offset = header_length(len(ifds))
+        image_offset = icc_offset + len(icc)
+
+        ifd_args = [(arg() if callable(arg) else arg) for ifd in ifds for arg in ifd]
         tiff_header = struct.pack(
-            tiff_header_struct,
+            (tiff_header_struct + ifd_struct[1:] * len(ifds) + 'L'),
             b'II',  # Byte order indication: Little endian
             42,  # Version number (always 42)
             8,  # Offset to first IFD
-            8,  # Number of tags in IFD
-            256, 4, 1, self.width,  # ImageWidth, LONG, 1, width
-            257, 4, 1, self.height,  # ImageLength, LONG, 1, length
-            258, 3, 1, 1,  # BitsPerSample, SHORT, 1, 1
-            259, 3, 1, ccitt_group,  # Compression, SHORT, 1, 4 = CCITT Group 4 fax encoding
-            262, 3, 1, int(photometry),  # Thresholding, SHORT, 1, 0 = WhiteIsZero
-            273, 4, 1, struct.calcsize(tiff_header_struct),  # StripOffsets, LONG, 1, length of header
-            278, 4, 1, self.height,
-            279, 4, 1, img_size,  # StripByteCounts, LONG, 1, size of image
-            0  # last IFD
+            len(ifds),  # Number of tags in IFD
+            *ifd_args,
+            0,  # Last IFD
         )
-        # fmt: on
+
+        if icc:
+            tiff_header += icc
+
         return tiff_header
 
-    def show(self):
+    def show(self):  # pragma: no cover
         """Show the image however PIL wants to"""
         self.as_pil_image().show()
 
     def __repr__(self):
-        return '<pikepdf.PdfImage image mode={} size={}x{} at {}>'.format(
-            self.mode, self.width, self.height, hex(id(self))
+        return (
+            f'<pikepdf.PdfImage image mode={self.mode} '
+            f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
     def _repr_png_(self):
@@ -620,7 +714,16 @@ class PdfImage(PdfImageBase):
 class PdfJpxImage(PdfImage):
     def __init__(self, obj):
         super().__init__(obj)
-        self.pil = self.as_pil_image()
+        self._jpxpil = self.as_pil_image()
+
+    def __eq__(self, other):
+        if not isinstance(other, PdfImageBase):
+            return NotImplemented
+        return (
+            self.obj == other.obj
+            and isinstance(other, PdfJpxImage)
+            and self._jpxpil == other._jpxpil
+        )
 
     def _extract_direct(self, *, stream):
         data, filters = self._unstack_compression(
@@ -638,9 +741,9 @@ class PdfJpxImage(PdfImage):
         super_colorspaces = super()._colorspaces
         if super_colorspaces:
             return super_colorspaces
-        if self.pil.mode == 'L':
+        if self._jpxpil.mode == 'L':
             return ['/DeviceGray']
-        elif self.pil.mode == 'RGB':
+        elif self._jpxpil.mode == 'RGB':
             return ['/DeviceRGB']
         raise NotImplementedError('Complex JP2 colorspace')
 
@@ -660,8 +763,9 @@ class PdfJpxImage(PdfImage):
         return False
 
     def __repr__(self):
-        return '<pikepdf.PdfJpxImage JPEG2000 image mode={} size={}x{} at {}>'.format(
-            self.mode, self.width, self.height, hex(id(self))
+        return (
+            f'<pikepdf.PdfJpxImage JPEG2000 image mode={self.mode} '
+            f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
 
@@ -688,6 +792,7 @@ class PdfInlineImage(PdfImageBase):
         b'/CCF': b'/CCITTFaxDecode',
         b'/DCT': b'/DCTDecode',
     }
+    REVERSE_ABBREVS = {v: k for k, v in ABBREVS.items()}
 
     def __init__(self, *, image_data, image_object: tuple):
         """
@@ -704,7 +809,9 @@ class PdfInlineImage(PdfImageBase):
         self._data = image_data
         self._image_object = image_object
 
-        reparse = b' '.join(self._unparse_obj(obj) for obj in image_object)
+        reparse = b' '.join(
+            self._unparse_obj(obj, remap_names=self.ABBREVS) for obj in image_object
+        )
         try:
             reparsed_obj = Object.parse(b'<< ' + reparse + b' >>')
         except PdfError as e:
@@ -712,13 +819,25 @@ class PdfInlineImage(PdfImageBase):
         self.obj = reparsed_obj
         self.pil = None
 
+    def __eq__(self, other):
+        if not isinstance(other, PdfImageBase):
+            return NotImplemented
+        return (
+            self.obj == other.obj
+            and isinstance(other, PdfInlineImage)
+            and (
+                self._data._inline_image_raw_bytes()
+                == other._data._inline_image_raw_bytes()
+            )
+        )
+
     @classmethod
-    def _unparse_obj(cls, obj):
+    def _unparse_obj(cls, obj, remap_names):
         if isinstance(obj, Object):
             if isinstance(obj, Name):
                 name = obj.unparse(resolved=True)
                 assert isinstance(name, bytes)
-                return cls.ABBREVS.get(name, name)
+                return remap_names.get(name, name)
             else:
                 return obj.unparse(resolved=True)
         elif isinstance(obj, bool):
@@ -732,35 +851,41 @@ class PdfInlineImage(PdfImageBase):
         return metadata_from_obj(self.obj, name, type_, default)
 
     def unparse(self):
-        tokens = []
-        tokens.append(b'BI')
-        metadata = []
-        for metadata_obj in self._image_object:
-            unparsed = self._unparse_obj(metadata_obj)
-            assert isinstance(unparsed, bytes)
-            metadata.append(unparsed)
-        tokens.append(b' '.join(metadata))
-        tokens.append(b'ID')
-        tokens.append(self._data._inline_image_raw_bytes())
-        tokens.append(b'EI')
-        return b'\n'.join(tokens)
+        def metadata_tokens():
+            for metadata_obj in self._image_object:
+                unparsed = self._unparse_obj(
+                    metadata_obj, remap_names=self.REVERSE_ABBREVS
+                )
+                assert isinstance(unparsed, bytes)
+                yield unparsed
+
+        def inline_image_tokens():
+            yield b'BI\n'
+            yield b' '.join(m for m in metadata_tokens())
+            yield b'\nID\n'
+            yield self._data._inline_image_raw_bytes()
+            yield b'EI'
+
+        return b''.join(inline_image_tokens())
 
     @property
     def is_inline(self):
         return True
 
     @property
-    def icc(self):
-        raise ValueError("Inline images may not have ICC profiles")
+    def icc(self):  # pragma: no cover
+        raise InvalidPdfImageError(
+            "Inline images with ICC profiles are not supported in the PDF specification"
+        )
 
     def __repr__(self):
-        mode = '?'
         try:
             mode = self.mode
-        except Exception:
-            pass
-        return '<pikepdf.PdfInlineImage image mode={} size={}x{} at {}>'.format(
-            mode, self.width, self.height, hex(id(self))
+        except NotImplementedError:
+            mode = '?'
+        return (
+            f'<pikepdf.PdfInlineImage image mode={mode} '
+            f'size={self.width}x{self.height} at {hex(id(self))}>'
         )
 
     def as_pil_image(self):
